@@ -40,12 +40,17 @@ defaults to L<Font::TTF::Table>. The current tables which are supported are:
     table       Font::TTF::Table      - for unknown tables
     EBDT        Font::TTF::EBDT
     EBLC        Font::TTF::EBLC
+    Feat        Font::TTF::GrFeat
     GDEF        Font::TTF::GDEF
     GPOS        Font::TTF::GPOS
     GSUB        Font::TTF::GSUB
+    Glat        Font::TTF::Glat
+    Gloc        Font::TTF::Gloc
     LTSH        Font::TTF::LTSH
     OS/2        Font::TTF::OS_2
     PCLT        Font::TTF::PCLT
+    Sill        Font::TTF::Sill
+    Silf        Font::TTF::Silf
     bsln        Font::TTF::Bsln
     cmap        Font::TTF::Cmap       - see also Font::TTF::OldCmap
     cvt         Font::TTF::Cvt_
@@ -68,18 +73,20 @@ defaults to L<Font::TTF::Table>. The current tables which are supported are:
     prop        Font::TTF::Prop
     vhea        Font::TTF::Vhea
     vmtx        Font::TTF::Vmtx
+    DSIG        FONT::TTF::DSIG
 
 Links are:
 
 L<Font::TTF::Table> 
-L<Font::TTF::EBDT> L<Font::TTF::EBLC>
-L<Font::TTF::GDEF> L<Font::TTF::GPOS> L<Font::TTF::GSUB> L<Font::TTF::LTSH>
-L<Font::TTF::OS_2> L<Font::TTF::PCLT> L<Font::TTF::Bsln> L<Font::TTF::Cmap> L<Font::TTF::Cvt_>
+L<Font::TTF::EBDT> L<Font::TTF::EBLC> L<Font::TTF::GrFeat>
+L<Font::TTF::GDEF> L<Font::TTF::GPOS> L<Font::TTF::GSUB> L<Font::TTF::Glat> L<Font::TTF::Gloc> L<Font::TTF::LTSH>
+L<Font::TTF::OS_2> L<Font::TTF::PCLT> L<Font::TTF::Sill> L<Font::TTF::Silf> L<Font::TTF::Bsln> L<Font::TTF::Cmap> L<Font::TTF::Cvt_>
 L<Font::TTF::Fdsc> L<Font::TTF::Feat> L<Font::TTF::Fmtx> L<Font::TTF::Fpgm> L<Font::TTF::Glyf>
 L<Font::TTF::Hdmx> L<Font::TTF::Head> L<Font::TTF::Hhea> L<Font::TTF::Hmtx> L<Font::TTF::Kern>
 L<Font::TTF::Loca> L<Font::TTF::Maxp> L<Font::TTF::Mort> L<Font::TTF::Name> L<Font::TTF::Post>
 L<Font::TTF::Prep> L<Font::TTF::Prop> L<Font::TTF::Vhea> L<Font::TTF::Vmtx> L<Font::TTF::OldCmap>
 L<Font::TTF::Glyph> L<Font::TTF::AATKern> L<Font::TTF::OldMort>
+L<Font::TTF::DSIG>
 
 
 =head1 INSTANCE VARIABLES
@@ -95,6 +102,18 @@ This is used during output to disable the creation of the file checksum in the
 head table. For example, during DSIG table creation, this flag will be set to
 ensure that the file checksum is left at zero.
 
+=item noharmony
+
+If set, do not harmonize the script and lang trees of GPOS and GSUB tables. See L<Font::TTF::Ttopen> for more info.
+
+=item nocompress
+
+Is the default value controlling WOFF output table compression. If undef, all tables will be compressed if there is 
+a size benefit in doing so. 
+It may be set to an array of tagnames naming tables that should not be compressed, or to a scalar integer specifying a 
+table size threshold below which tables will not be compressed. 
+Note that individual L<Font::TTF::Table> objects may override this default. See L<Font::TTF::Table> for more info.
+
 =item fname (R)
 
 Contains the filename of the font which this object was read from.
@@ -108,6 +127,10 @@ The file handle which reflects the source file for this font.
 Contains the offset from the beginning of the read file of this particular
 font directory, thus providing support for TrueType Collections.
 
+=item WOFF
+
+Contains a reference to a C<Font::TTF::Woff> object.
+
 =back
 
 =head1 METHODS
@@ -122,7 +145,11 @@ use Symbol();
 
 require 5.004;
 
-$VERSION = 0.37;    # MJPH       7-OCT-2005     Force hhea update if dirty, give more OS/2 stuff in update
+my $havezlib = eval {require Compress::Zlib};
+
+$VERSION = 0.39;    # MJPH       2-FEB-2008     Add DSIG table
+# $VERSION = 0.38;    # MJPH       2-FEB-2008     Add Sill table
+# $VERSION = 0.37;    # MJPH       7-OCT-2005     Force hhea update if dirty, give more OS/2 stuff in update
 # $VERSION = 0.36;    # MJPH      19-AUG-2005     Change cmap::reverse api to be opts based
 # $VERSION = 0.35;    # MJPH       4-MAY-2004     Various fixes to OpenType stuff, separate off scripts
 # $VERSION = 0.34;    # MJPH      22-MAY-2003     Update PSNames to latest AGL
@@ -166,14 +193,22 @@ $VERSION = 0.37;    # MJPH       7-OCT-2005     Force hhea update if dirty, give
 
 %tables = (
         'table' => 'Font::TTF::Table',
+        'DSIG' => 'Font::TTF::DSIG',
         'EBDT' => 'Font::TTF::EBDT',
         'EBLC' => 'Font::TTF::EBLC',
+        'Feat' => 'Font::TTF::GrFeat',
         'GDEF' => 'Font::TTF::GDEF',
+        'Glat' => 'Font::TTF::Glat',
+        'Gloc' => 'Font::TTF::Gloc',
         'GPOS' => 'Font::TTF::GPOS',
         'GSUB' => 'Font::TTF::GSUB',
+        'Glat' => 'Font::TTF::Glat',
+        'Gloc' => 'Font::TTF::Gloc',
         'LTSH' => 'Font::TTF::LTSH',
         'OS/2' => 'Font::TTF::OS_2',
         'PCLT' => 'Font::TTF::PCLT',
+        'Sill' => 'Font::TTF::Sill',
+        'Silf' => 'Font::TTF::Silf',
         'bsln' => 'Font::TTF::Bsln',
         'cmap' => 'Font::TTF::Cmap',
         'cvt ' => 'Font::TTF::Cvt_',
@@ -309,7 +344,7 @@ sub open
 
 =head2 $f->read
 
-Reads a Truetype font directory starting from the current location in the file.
+Reads a Truetype font directory starting from location C<$self->{' OFFSET'}> in the file.
 This has been separated from the C<open> function to allow support for embedded
 TTFs for example in TTCs. Also reads the C<head> and C<maxp> tables immediately.
 
@@ -320,17 +355,90 @@ sub read
     my ($self) = @_;
     my ($fh) = $self->{' INFILE'};
     my ($dat, $i, $ver, $dir_num, $type, $name, $check, $off, $len, $t);
+    my ($iswoff, $woffLength, $sfntSize, $zlen);    # needed for WOFF files
 
     $fh->seek($self->{' OFFSET'}, 0);
-    $fh->read($dat, 12);
-    ($ver, $dir_num) = unpack("Nn", $dat);
-    $ver == 1 << 16 || $ver == unpack('N', 'OTTO') || $ver == 0x74727565 or return undef;  # support Mac sfnts
+    $fh->read($dat, 4);
+    $ver = unpack("N", $dat);
+    $iswoff = ($ver == unpack('N', 'wOFF'));
+    if ($iswoff)
+    {
+        require Font::TTF::Woff;
+        my $woff = Font::TTF::Woff->new(PARENT  => $self);
+        $fh->read($dat, 32);
+        ($ver, $woffLength, $dir_num, undef, $sfntSize, $woff->{'majorVersion'}, $woff->{'minorVersion'}, 
+            $off, $zlen, $len) = unpack('NNnnNnnNNN', $dat);
+        # TODO: According to WOFF spec we should verify $woffLength and $sfntSize, and fail if the values are wrong.
+        if ($off)
+        {
+            # Font has metadata
+            if ($off + $zlen > $woffLength)
+            {
+                warn "invalid WOFF header in $self->{' fname'}: meta data beyond end.";
+                return undef;
+            }
+            require Font::TTF::Woff::MetaData;
+            $woff->{'metaData'} = Font::TTF::Woff::MetaData->new(
+                PARENT     => $woff,
+                INFILE     => $fh,
+                OFFSET     => $off,
+                LENGTH     => $len,
+                ZLENGTH    => $zlen);
+        }
+            
+        $fh->read($dat, 8);
+        ($off, $len) = unpack('NN', $dat);
+        if ($off)
+        {
+            # Font has private data
+            if ($off + $len > $woffLength)
+            {
+                warn "invalid WOFF header in $self->{' fname'}: private data beyond end.";
+                return undef;
+            }
+            require Font::TTF::Woff::PrivateData;
+            $woff->{'privateData'} = Font::TTF::Woff::PrivateData->new(
+                PARENT     => $woff,
+                INFILE     => $fh,
+                OFFSET     => $off,
+                LENGTH     => $len);
+        } 
+        
+        $self->{' WOFF'} = $woff;
+    }
+    else
+    {
+        $fh->read($dat, 8);
+        $dir_num = unpack("n", $dat);
+    }
+    
+    $ver == 1 << 16                 # TrueType outlines
+    || $ver == unpack('N', 'OTTO')  # 0x4F54544F CFF outlines
+    || $ver == unpack('N', 'true')  # 0x74727565 Mac sfnts
+    or return undef;            # else unrecognized type
+    
     
     for ($i = 0; $i < $dir_num; $i++)
     {
-        $fh->read($dat, 16) || die "Reading table entry";
-        ($name, $check, $off, $len) = unpack("a4NNN", $dat);
-        $self->{$name} = $self->{' PARENT'}->find($self, $name, $check, $off, $len) && next
+        $fh->read($dat, $iswoff ? 20 : 16) || die "Reading table entry";
+        if ($iswoff)
+        {
+            ($name, $off, $zlen, $len, $check) = unpack("a4NNNN", $dat);
+            if ($off + $zlen > $woffLength || $zlen > $len)
+            {
+                my $err;
+                $err = "Offset + compressed length > total length. " if $off + $zlen > $woffLength;
+                $err = "Compressed length > uncompressed length. " if $zlen > $len;
+                warn "invalid WOFF '$name' table in $self->{' fname'}: $err\n";
+                return undef;
+            }
+        }
+        else
+        {
+            ($name, $check, $off, $len) = unpack("a4NNN", $dat);
+            $zlen = $len;
+        }
+        $self->{$name} = $self->{' PARENT'}->find($self, $name, $check, $off, $len) and next
                 if (defined $self->{' PARENT'});
         $type = $tables{$name} || 'Font::TTF::Table';
         $t = $type;
@@ -344,6 +452,7 @@ sub read
                                     INFILE  => $fh,
                                     OFFSET  => $off,
                                     LENGTH  => $len,
+                                    ZLENGTH => $zlen,
                                     CSUM    => $check);
     }
     
@@ -357,7 +466,7 @@ sub read
 =head2 $f->out($fname [, @tablelist])
 
 Writes a TTF file consisting of the tables in tablelist. The list is checked to
-ensure that only tables that exist are output. (This means that you can't have
+ensure that only tables that exist are output. (This means that you cannot have
 non table information stored in the font object with key length of exactly 4)
 
 In many cases the user simply wants to output all the tables in alphabetical order.
@@ -378,6 +487,8 @@ sub out
     my (%dir, $k, $mloc, $count);
     my ($csum, $lsum, $msum, $loc, $oldloc, $len, $shift);
 
+    my ($iswoff); # , $woffLength, $sfntSize, $zlen);   # needed for WOFF files
+
     unless (ref($fname))
     {
         $fh = IO::File->new("+>$fname") || return warn("Unable to open $fname for writing"), undef;
@@ -405,10 +516,16 @@ sub out
     $numTables = $#tlist + 1;
     $numTables++ if ($self->{' wantsig'});
     
-    ($numTables, $sRange, $eSel, $shift) = Font::TTF::Utils::TTF_bininfo($numTables, 16);
-    $dat = pack("Nnnnn", 1 << 16, $numTables, $sRange, $eSel, $shift);
-    $fh->print($dat);
-    $msum = unpack("%32N*", $dat);
+    if ($iswoff)
+    {
+    }
+    else
+    {
+        ($numTables, $sRange, $eSel, $shift) = Font::TTF::Utils::TTF_bininfo($numTables, 16);
+        $dat = pack("Nnnnn", 1 << 16, $numTables, $sRange, $eSel, $shift);
+        $fh->print($dat);
+        $msum = unpack("%32N*", $dat);
+    }
 
 # reserve place holders for each directory entry
     foreach $k (@tlist)
@@ -429,9 +546,53 @@ sub out
     foreach $k (@tlist)
     {
         $oldloc = $loc;
-        $self->{$k}->out($fh);
-        $loc = $fh->tell();
-        $len = $loc - $oldloc;
+        if ($iswoff && $havezlib &&
+            # output font is WOFF -- should we try to compress this table?
+            exists ($self->{$k}->{' nocompress'}) ? $self->{$k}->{' nocompress'} != -1 :
+            ref($self->{' nocompress'}) eq 'ARRAY' ? !exists($self->{' nocompress'}{$k}) :
+            ref($self->{' nocompress'}) eq 'SCALAR' && $self->{' nocompress'} != -1)
+        {
+            # Yes -- we may want to compress this table.
+            # Create string file handle to hold uncompressed table
+            my $dat;
+            my $fh2 = IO::String->new($dat);
+            binmode $fh2;
+            $self->{$k}->out($fh2);
+            $len = $fh2->tell();
+            close $fh2;
+            
+            # Is table long enough to try compression?
+            unless (
+                exists ($self->{$k}->{' nocompress'}) && $len <= $self->{$k}->{' nocompress'} ||
+                ref($self->{' nocompress'}) eq 'SCALAR' && $len <= $self->{' nocompress'})
+            {
+                # Yes -- so compress and check lengths:
+                my $zdat = Compress::Zlib::compress($dat);
+                my $zlen = bytes::length($zdat);
+                if ($zlen < $len)
+                {
+                    # write the compressed $zdat
+                    
+                }
+                else
+                {
+                    # write the uncompressed $dat
+                }
+            }
+            else
+            {
+                # write uncompressed $dat
+            }
+            
+            
+        }
+        else
+        {
+            # Output table normally
+            $self->{$k}->out($fh);
+            $loc = $fh->tell();
+            $len = $loc - $oldloc;
+        }
         if ($loc & 3)
         {
             $fh->print(substr("\000" x 4, $loc & 3));
@@ -630,6 +791,7 @@ order as per the order in the directory:
 
 May optionally take a list of table names in which case func is called
 for each of them in the given order.
+
 =cut
 
 sub tables_do
@@ -712,10 +874,14 @@ somewhat spurious results for this table type.
 
 Only type 0 & type 2 tables are supported (type 1 & type 3 yet to come).
 
-=item TTC
+=item TTC and WOFF
 
 The current Font::TTF::Font::out method does not support the writing of TrueType
-Collections.
+Collections or WOFF files.
+
+=item DSIG
+
+Haven't figured out how to correctly calculate and output digital signature (DSIG) table
 
 =back
 
@@ -745,14 +911,17 @@ For more details see the appropriate class files.
 
 =head1 AUTHOR
 
-Martin Hosken Martin_Hosken@sil.org
+Martin Hosken L<http://scripts.sil.org/FontUtils>.
 
-Copyright Martin Hosken 1998.
 
-No warranty or expression of effectiveness, least of all regarding anyone's
-safety, is implied in this software or documentation.
+=head1 LICENSING
 
-=head2 Licensing
+Copyright (c) 1998-2016, SIL International (http://www.sil.org) 
 
-The Perl TTF module is licensed under the Perl Artistic License.
+This module is released under the terms of the Artistic License 2.0. 
+For details, see the full text of the license in the file LICENSE.
+
+
+
+=cut
 
